@@ -6,28 +6,46 @@ from scipy.stats import percentileofscore
 from scipy.stats.distributions import chi2
 from scipy.special import expit, xlog1py, xlogy
 from tqdm import tqdm
+from numba import njit
 from .MakeTable import MakeTwoByTwoTable
 from .Marginal import calculateMarginals
 
+@njit
 def kbcom(k, b):
     return -1 if k == b else 0
 
-def itemPb2(q, s, k, b, xVari, itemi, linear = False):
-    vS = np.dot(xVari, itemi) + q + s if linear else expit(np.dot(xVari, itemi) + q + s)
-    return  xlogy(1, vS + (vS - 1) * kbcom(k, b)) + xlog1py(k, -1*vS)
+@njit
+def itemPb2(q, s, k, b, xVari, itemi):
+    return ( np.dot(np.ascontiguousarray(xVari), np.ascontiguousarray(itemi)) + q + s, k, kbcom(k, b))
 
-def itemPb(q, s, k, b, linear = False):
-    vS = q + s if linear else expit(q + s)
-    return  xlogy(1, vS + (vS - 1) * kbcom(k, b)) + xlog1py(k, -1*vS)
+@njit
+def itemPb(q, s, k, b):
+    return ( q + s, k, kbcom(k, b))
+
+@njit
+def itemLoop2(x, data, cols = 0):
+    return [ itemPb2(x[int(item[2])], x[int(item[1])], item[0], item[3], x[-cols:], item[-cols:]) for item in data ]
+
+@njit
+def itemLoop(x, data):
+    return [ itemPb(x[int(item[2])], x[int(item[1])], item[0], item[3]) for item in data ]
+
+@njit
+def itemLoopRestricted(x, data):
+    return [ ( x[0], item[0], kbcom(item[0], item[3])) for item in data ]
 
 def opFunction(x, data, linear = False, cols = 0):
     if cols > 0:
-        return -1.0 * (sum(( itemPb2(x[item[2]], x[item[1]], item[0], item[3], x[-cols:], item[-cols:], linear) for item in data )))
+        dem = np.array(itemLoop2(x, data, cols))
     else:
-        return -1.0 * (sum(( itemPb(x[item[2]], x[item[1]], item[0], item[3], linear) for item in data )))
+        dem = np.array(itemLoop(x, data))
+    vS = dem[:,0] if linear else expit(dem[:,0])
+    return -1.0 * np.sum(xlogy(1,  vS + (vS - 1) * dem[:,2]) + xlog1py(dem[:,1], -vS))
 
 def opRestricted (x, data, linear = False):
-    return -1.0 * (sum(( itemPb(x[0], 0, item[0], item[3], linear) for item in data )))
+    dem = np.array(itemLoopRestricted(x, data))
+    vS = dem[:,0] if linear else expit(dem[:,0])
+    return -1.0 * np.sum(xlogy(1,  vS + (vS - 1) * dem[:,2]) + xlog1py(dem[:,1], -vS))
 
 def solve(dataset, summary = True, linear = False, columns = None):
     studentCode, uniqueStudents = pd.factorize(dataset['student'])
@@ -42,7 +60,7 @@ def solve(dataset, summary = True, linear = False, columns = None):
         bounds = [(0, 1)] * (uniqueStudents.size + uniqueQuestion.size + len(columns))
     else:
         bounds = None
-    minValue = minimize(opFunction, [1/(2*(1+dataset['k'].mean()))]*(len(smap) + len(columns)), args=(data, linear, len(columns)), method='Powell', bounds=bounds)
+    minValue = minimize(opFunction, np.array([1/(2*(1+dataset['k'].mean()))]*(len(smap) + len(columns)), np.dtype(float)), args=(np.array(data, np.dtype(float)), linear, len(columns)), method='Powell', bounds=bounds)
     if minValue.success:
         estX = minValue.x.flatten().tolist()
         varNames = smap + columns
@@ -72,7 +90,7 @@ def solve(dataset, summary = True, linear = False, columns = None):
             boundsSingle = [(0, 1)]
         else:
             boundsSingle = None
-        minRestricted = minimize(opRestricted, [1/(1+dataset['k'].mean()),], args=(data, linear), method='Powell', bounds=boundsSingle)
+        minRestricted = minimize(opRestricted, np.array([1/(1+dataset['k'].mean()),], np.dtype(float)), args=(np.array(data, np.dtype(float)), linear), method='Powell', bounds=boundsSingle)
         if minRestricted.success:
             d['McFadden'] = 1 - minValue.fun/minRestricted.fun
             d['LR'] = -2*(-1*minRestricted.fun+minValue.fun)
@@ -147,7 +165,7 @@ def getResults(dataset: pd.DataFrame,c=0.025, rubric=False, n=1000, linear=False
     linear: bool
         Uses a simple linear combination of the rubric and student items instead of a sigmoid function.  Defaults to False.
     columns: list
-        A list of column names to include in the model.  Defaults to None.
+        A list of column names to include in the model.  The column names cannot be in common with any of the rubric row identifiers.  Defaults to None.
 
     Returns:
         Tuple:
@@ -231,7 +249,7 @@ def DisplayResults(dataset: pd.DataFrame,c=0.025, rubric=False, n=1000, linear=F
     linear: bool
         Uses a simple linear combination of the rubric and student items instead of a sigmoid function.  Defaults to False.
     columns: list
-        A list of column names to include in the model.  Defaults to None.
+        A list of column names to include in the model. The column names cannot be in common with any of the rubric row identifiers. Defaults to None.
 
     Returns:
         Tuple:
@@ -314,7 +332,7 @@ def SaveResults(dataset: pd.DataFrame,c=0.025, rubric=False, n=1000, linear=Fals
     outputFile : str
         File name/path for the summary output results.  Defaults to 'output.csv'.
     columns: list
-        A list of column names to include in the model.  Defaults to None.
+        A list of column names to include in the model. The column names cannot be in common with any of the rubric row identifiers. Defaults to None.
 
     Returns:
         Tuple:
